import sys

sys.path.insert(0, './modules')

import os
from typing import *
from flask import Flask
import settings
from modules.h5ceric import extract_dataset, plot_dispatcher
from modules.h5decorators import nocache
import flask

if not os.path.exists("./data"):
    try:
        dataDir = "data"
        parentDir = "."
        toBeCreated = os.path.join(parentDir, dataDir)
        os.mkdir(toBeCreated, 0o755)
        print("data folder created")
    except Exception as e:
        print("Could not find nor create data directory: ", e)

deploymentType = settings.deploy

app = Flask(__name__)


########################## helper function ##################
def split_url(path: str) -> Tuple[str, str]:
    """helper function to split a flask path

    This helper function accept as input a string, in the format "<file_name>,dataset=<dataset_name>",
    and split it to return the two variables in a tuple

    :param path: the path accessed in flask
    :return: a tuple (file_name, dataset_name) with
    """
    path_array = path.split(",dataset=")

    file_name = path_array[0]  # path of the file folder
    dataset_name = path_array[-1]  # filename
    return file_name, dataset_name


############################# homepage ###########################
@app.route('/')
@nocache
def hello() -> Any:
    """ main flask web page renderer

    Called when accessing the root URL of the website, in returns the index.html page, populating it.

    :return: it renders the template templates/index.html
    """

    print("start folder: ", settings.fileFolder[deploymentType])
    print("deployment type: ", settings.deploy)
    from modules.jsTreeCeric import JSON_from_folder
    folder_dictionary = JSON_from_folder(settings.fileFolder[deploymentType])
    # print("folder dict : ", folder_dictionary)
    import json
    folder_tree = json.dumps(folder_dictionary)

    h5_tree = json.dumps([])
    from flask import render_template
    return render_template("index.html", treeJsonFromFlask=folder_tree, h5Tree=h5_tree)


########################### hdf5 tree ############################
@app.route('/openhdf5/<path:path>')
@nocache
def openhdf5(path: str) -> str:
    """ Route to access hdf5 file content

    Opens a hdf5 file, get a dictionary of groups and datasets, converts it in
    a JSON ready to be read from a jsTree call.

    :param path: a URL route path
    :return: a JSON, formatted as jsTree input, that list the content of the hdf5 file located in <path>
    """
    from modules.h5ceric import list_datasets

    import json
    h5_tree = json.dumps(list_datasets(settings.fileFolder[deploymentType] + path))
    return h5_tree


######################### bokeh visualization ######################
@app.route('/visualize/')
@nocache
def send_js() -> Any:
    """plots a hdf5 dataset

    :return: plot the dataset with Bokeh, and renders the template template.html with the plot
    """
    from flask import render_template, request
    from modules.h5ceric import DatabaseNotFound

    # get dataset as np.array

    filename = request.args.get('filename', None)
    ds_name = request.args.get('dataset', None)
    z = request.args.get('z', None)

    try:
        arr = extract_dataset(settings.fileFolder[deploymentType], filename, ds_name)
    except OSError:
        return render_template("missing_file.html")
    except DatabaseNotFound:
        return render_template("missing_dataset.html")

    # create the plot
    import numpy

    if isinstance(arr, numpy.ndarray):
        p = plot_dispatcher(arr, ds_name, z)

        # forced garbage collection, because arr isn't automatically destroyed,
        # and this keeps using more and more memory
        import gc
        gc.collect()

        # split out plot script
        from bokeh.embed import components
        script, div = components(p)

        # embed in html
        return render_template('template.html', script=script, div=div)


######################### get file info ######################
@app.route('/file_info/<path:path>')
@nocache
def file_info(path: str) -> flask.Response:
    """
    return mimetype and URL of a file
    :param path: file path
    :return: a JSON with "mime_type" and "url" fields.
    """

    from mimetypes import guess_type
    from flask import Response
    import json
    info = json.dumps(
        {"mime_type": guess_type(settings.fileFolder[deploymentType] + path)[0],
         "url": "/get_file/" + path}
    )

    response = Response(info)
    response.headers['Content-Type'] = 'application/json'
    return response


######################### read file content ######################
@app.route('/get_file/<path:path>')
@nocache
def get_file(path: str) -> Any:
    """
    return an header with correct mime_type and file content.

    e.g.  https://stackoverflow.com/questions/11017466/flask-to-return-image-stored-in-database

    :param path: file path
    :return: flask.send_file() instance
    """

    from flask import send_file
    from mimetypes import guess_type

    fullpath = settings.fileFolder[deploymentType] + path
    mimetype = guess_type(fullpath)[0]

    return send_file(fullpath,
                     attachment_filename=path.split('/')[-1],  # 'file-png'
                     mimetype=mimetype)


######################### read file content ######################
@app.route('/display_report/')
def display_report() -> str:
    """
    receiving a request with all the elements to be visualized, compiles the HTML of
    the report
    :return: HTML of the report
    """

    import json
    import h5py
    from flask import request
    report = request.args.get('report', None)
    report_data = json.loads(report)
    html = """<!DOCTYPE html>\n
                <html lang="en">\n
                <head>\n
                <title>H5Nuvola report</title>\n
                </head>\n
                <body>\n
                <h1>h5nuvola report</h1>\n
                <!-- bootstrap -->\n
                <link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.4.1/css/bootstrap.min.css" integrity="sha384-Vkoo8x4CGsO3+Hhxv8T/Q5PaXtkKtu6ug5TOeNV6gBiFeWPGFN9MuhOf23Q9Ifjh" crossorigin="anonymous">\n
    
            """
    for filename, content in report_data.items():
        html += '<h4>Filename:{}</h4>'.format(filename)
        for item in content:

            # get type of item["dataset"] in item["filename"]
            if filename.split(".")[-1] in ["nxs", "h5", "hdf5", "nexus"]:

                with h5py.File(settings.fileFolder[deploymentType] + "/" + filename, 'r') as f:
                    # dispatch plotting javascript
                    if len(f[item['dataset']].shape) == 0:

                        html += '<ul>\n'
                        html += '<li><strong>Single value:</strong> ' + str(item["dataset"]) + '</li>\n'
                        value = f[item["dataset"]].value
                        html += '<li><strong>Value:</strong> ' + str(value) + '</li>\n'
                        html += '</ul>\n'

                    # if it's a 2D dataset...
                    elif len(f[item['dataset']].shape) == 2:
                        html += '<ul>\n'
                        html += '<li><strong>2D Dataset:</strong> ' + str(item["dataset"]) + '</li>\n'
                        html += '<li><strong>Image:</strong><br>\n<img src="' + str(item["image"]).replace(" ",
                                                                                                           "+") + '"></li>\n'
                        html += '</ul>\n'

                    # if it's a 3D dataset, get the current slice.
                    elif len(f[item['dataset']].shape) == 3:
                        html += '<ul>\n'
                        html += '<li><strong>3D Dataset:</strong> ' + str(item["dataset"]) + '</li>\n'
                        html += '<li><strong>Z:</strong> ' + str(item["z"]) + '</li>\n'
                        html += '<li><strong>Image:</strong><br>\n<img src="' + str(item["image"]).replace(" ",
                                                                                                           "+") + '"></li>\n'
                        html += '</ul>\n'

            else:
                display_code = ''
                if "image" in item["mime"]:
                    display_code = '<img class="img-responsive" style="height: 75vh; max-width: 100%;" src="' + item[
                        "url"] + '" >'
                elif "video" in item["mime"]:
                    display_code = '<video class="img-responsive" style="height: 75vh; max-width: 100%;" video-element" controls><source type="' + \
                                  item['mime'] + '" src="' + item['url'] + '"></video>'
                elif "text" in item["mime"]:
                    display_code = '<iframe src="' + item["url"] + '" style="height: 75vh; max-width: 100%;">'
                html += '<ul>\n'
                html += display_code

                html += '</ul>\n'

    html += '</body>\n</html>\n'
    # embed in html
    return html


# With debug=True, Flask server will auto-reload
# when there are code changes
if __name__ == '__main__':
    app.run(host=settings.host[deploymentType],
            port=settings.port,
            debug=settings.debug)
