__name__ = "h5ceric"

import collections
from typing import Dict, Any, Union

import h5py
import numpy


def traverse_datasets(hdf_file: h5py.File):
    """ Generator for hdf5 folder traversal

    Yields a generator that iterates the datasets included in an hdf5 file.

    :param hdf_file: a string with the path of an hdf5 file
    :return a generator from an iteration of a hdf5 file
    """
    for path, _ in h5py_dataset_iterator(hdf_file):
        yield path


def list_datasets(name: str, prefix: str = '') -> Dict:
    """List all datasets in an hdf5 file

    :param name: path of the file
    :param prefix: used for recursive call. don't set it.
    :return: a list of all the datasets in the required file
    """
    import settings  # for icons
    output = {"core": {"data": []}}
    d = []

    def iterate_hdf5(h5file: h5py.File) -> Any:
        """ hdf5 file iterator

        Iterates an hdf5 file, returning a JSON formatted to be used in JSTree visualization

        :param h5file: hdf5 file
        :returns output: a dict to be converted in JSON for JSTree
        """
        partial_tree = []
        for keys in list(h5file.keys()):
            if isinstance(h5file[keys], str):
                partial_tree.append(h5file[keys])

            if isinstance(h5file[keys], h5py.Dataset):
                partial_tree.append({'text': h5file[keys].name,
                                     'icon': settings.datasetIcon})

            elif isinstance(h5file[keys], h5py.Group):
                partial_tree.append({'text': h5file[keys].name,
                                     'children': iterate_hdf5(h5file[keys]),
                                     'icon': settings.hdf5Icon})
        return partial_tree

    if isinstance(name, str):
        try:
            with h5py.File(name, "r") as h5:
                d = iterate_hdf5(h5)
        except OSError:
            print("could not open hdf5 file :  ", name)

    output['core'] = {'data': d}

    return output


def dict_datasets(name: str) -> None:
    """Prints hdf5 datasets.

     Honestly, h5glance.H5Glance works better in notebooks.

     :param name: path of the hdf5 file
    """
    with h5py.File(name, "r") as f:
        print("{: >20}{: >20}{: >20}".format("Path", "Size", "dType") + "\n" + "-" * 60)
        print()
        for dataset in traverse_datasets(f):
            text = (dataset, str(f[dataset].shape), str(f[dataset].dtype))
            print("{: >20}{: >20}{: >20}".format(*text))


def h5py_dataset_iterator(file: Union[h5py.File, h5py.Group], prefix: object = '') -> collections.Iterable:
    """Hdf5 dataset iterator

    Recursively iterates a dataset, returns an iterator over all datasets

    :param prefix: used only for recursive calls
    :param file: path of a valid hdf5 file
    :return an iterator over file's datasets
    """

    for key in file.keys():
        items = file[key]
        path = f'{prefix}/{key}'
        if isinstance(items, h5py.Dataset):  # test for dataset
            yield path, items
        elif isinstance(items, h5py.Group):  # test for group (go down)
            yield from h5py_dataset_iterator(items, path)


class DatabaseNotFound(Exception):
    """generic exception for trying to access a non-existent dataset in a hfd5 file"""
    pass


def extract_dataset(folder: str, filename: str, required_dataset: str):
    """Retrieve numpy arrays from hdf5 datasets.
    
    :param folder: use settings.folder, it's the absolute path of the hdf5 folder  
    :param filename: name and extension of the hdf5 file
    :param required_dataset: full path of the required dataset: "/entry/.../..."
    :return: numpy.array with the required data
    """
    full_path = folder + filename
    import numpy as np

    with h5py.File(full_path, "r") as f:
        # print("opened file "+full_path)
        for dataset_name, dataset in h5py_dataset_iterator(f):
            # print("checking if "+dataset_name+" matches required dataset")
            if required_dataset == dataset_name:
                return np.array(dataset)
    raise OSError


def plot_dispatcher(arr: numpy.ndarray, ds_name: str, pos: int = 0):
    """Creates a bokeh plot from a numpy 1D or 2D array

    If array order > 2, a 2D slice of arr[0] will be considered.

    :param arr: numpy array, 1-2D
    :param ds_name: dataset name, will be used for title label
    :param pos: default slice
    :return: a bokeh plot
    """

    # plot single data
    if len(arr.shape) == 0:
        from hdf5BokehPlots import plot_single_values
        return plot_single_values(arr, ds_name)

    # plot 1D #

    if len(arr.shape) == 1:
        from hdf5BokehPlots import plot_1d_bokeh
        return plot_1d_bokeh(arr, ds_name)

    # plot 2D #
    if len(arr.shape) == 2:
        from math import isclose

        if isclose(arr.shape[0], arr.shape[1], rel_tol=0.3):
            from hdf5BokehPlots import plot_2d_bokeh
            return plot_2d_bokeh(arr, ds_name)
        else:
            from hdf5BokehPlots import plot_2d_series
            import numpy as np
            import settings
            if arr.shape[0] > settings.maxPointInHexPlot:
                arr = arr[np.random.choice(arr.shape[0], settings.maxPointInHexPlot, replace=False), :]
            return plot_2d_series(arr[:, 0], arr[:, 1], ds_name)

    # plot 3+D #
    # slice array if order > 3: this needs to be fixed for higher dimensions
    if len(arr.shape) >= 3:
        from hdf5BokehPlots import plot_3d_bokeh
        return plot_3d_bokeh(arr, ds_name, pos)
