import numpy


def plot_single_values(arr: numpy.ndarray, ds_name: str) -> dict:
    """
    create a canvas with a single value (will be showed in plot title).

    :param arr: a 0D numpy array
    :param ds_name: string, plot title
    :return: a bokeh figure
    """
    from bokeh.plotting import figure

    p = figure(title=ds_name + " = " + str(arr),
               toolbar_location="above",
               x_range=(0, 0), y_range=(0, 0),
               match_aspect=True,
               sizing_mode='stretch_both'
               )

    p.grid.grid_line_width = 0.0

    return {"plot": p}


def plot_1d_bokeh(arr: numpy.ndarray, ds_name: str) -> dict:
    """ creates a bokeh plot from a 1D time series

    creates a bokeh plot from a 1D time series

    :param arr: 1D numpy ndarray
    :param ds_name: string, plot title
    :return: a bokeh figure
    """
    from bokeh.plotting import figure
    from bokeh.models import DataRange1d
    import numpy  # linspace

    size_x = arr.shape[0]

    x = numpy.linspace(0, len(arr), size_x)

    x_range = DataRange1d(range_padding=0, range_padding_units='percent')

    # generate and return bokeh plot
    p = figure(title=ds_name,
               toolbar_location="above",
               x_range=x_range, y_range=(min(arr) * 1.05, max(arr) * 1.05),
               match_aspect=True,
               sizing_mode='stretch_both'
               )
    p.grid.grid_line_width = 0.2

    p.line(x, arr, color="navy", alpha=0.5)

    # hovertools section
    from bokeh.models import HoverTool

    # Format the tooltip
    tooltips = [
        ('Time', '@x{0.2f}'),
        ('Value: ', '@y')
    ]

    # Add the HoverTool to the figure
    p.add_tools(HoverTool(tooltips=tooltips))

    return {"plot": p}


def plot_2d_bokeh(arr: numpy.ndarray, ds_name: str) -> dict:
    """ creates a bokeh plot from a 2D figure

        creates a bokeh plot from a 2D figure

        :param arr: 2D numpy ndarray
        :param ds_name: string, plot title
        :return: a bokeh figure
        """
    from bokeh.plotting import figure
    from bokeh.models import DataRange1d

    size_x, size_y = arr.shape

    x_range = DataRange1d(range_padding=0, range_padding_units='percent')
    y_range = DataRange1d(range_padding=0, range_padding_units='percent')

    # generate and return bokeh plot
    p = figure(title=ds_name,
               toolbar_location="above",
               x_range=x_range, y_range=y_range,
               match_aspect=True,
               sizing_mode='stretch_both'
               )

    p.grid.grid_line_width = 0.2

    p.image(image=[arr], x=0, y=0, dw=size_x, dh=size_y,
            palette="Viridis256",
            level="image"
            )

    return {"plot": p}


def plot_2d_series(x: numpy.ndarray, y: numpy.ndarray, ds_name='') -> dict:
    """ creates a bokeh plot from a 2D figure

        :param y:  values of the series
        :param x: x axis of the series (time?)
        :param ds_name: title of the plot
        :return: a bokeh figure
        """

    import holoviews as hv

    hv.extension('bokeh')

    import numpy as np
    arr = np.stack((x, y), axis=0).T
    points = hv.Points(arr, title=ds_name)
    points = points.hist(num_bins=51, dimension=['x', 'y'])
    points.main.opts(width=500, height=500, size=0.8)
    return {"plot": hv.render(points, backend='bokeh')}


def plot_3d_bokeh(arr: numpy.ndarray, ds_name: str, y: int) -> dict:
    """ creates a bokeh plot from a 3D matrix

            creates a bokeh plot from a 3D matrix

            :param y: slice position
            :param arr: 3D numpy ndarray
            :param ds_name: string, plot title
            :return: a bokeh figure
            """

    if len(arr.shape) > 3:  # slice off excessive dimensions
        arr = arr[0]
        assert (len(arr.shape) == 3)
    y = int(y)
    size_x, size_y, size_z = arr.shape

    from numpy import rot90
    arr = rot90(arr, 3, (0, 2))

    if y > 0:
        slice_y = y
    else:
        slice_y = int(size_y / 4)

    from bokeh.models import DataRange1d
    x_range = DataRange1d(range_padding=0, range_padding_units='percent')
    y_range = DataRange1d(range_padding=0, range_padding_units='percent')

    # generate and return bokeh plot
    from bokeh.plotting import figure
    p = figure(title=ds_name,
               toolbar_location="above",
               x_range=x_range, y_range=y_range,
               match_aspect=True,
               sizing_mode='stretch_both'
               )

    p.grid.grid_line_width = 0.2

    from bokeh.models import ColorBar, LinearColorMapper
    from numpy import amax, amin
    from settings import plotPalette_3D

    color_mapper = LinearColorMapper(palette=plotPalette_3D,
                                     low=amin(arr),
                                     high=amax(arr))

    color_bar = ColorBar(color_mapper=color_mapper,
                         label_standoff=12,
                         border_line_color=None,
                         location=(0, 0))

    im = [arr[slice_y]]  # first default slice

    p.image(image=im, x=0, y=0, dw=size_x, dh=size_z,
            color_mapper=color_mapper,
            level="image"
            )

    from bokeh.models import Slider
    z_slider = Slider(start=0, end=size_y - 1, value=slice_y, step=1, title="Z")

    from bokeh.models import CustomJS
    callback = CustomJS(args=dict(slider=z_slider),
                        code="""
                            $("#addToRepButton").remove(); generateBokehPlot(slider.value);
                            """)

    z_slider.js_on_change('value', callback)

    p.add_layout(color_bar, 'right')

    return {"plot": p, "slider": z_slider}
